@@ -1,41 +1,39 @@
-import os
-import signal
 import asyncio
-import logging
-import pathlib
-import warnings
-import threading
 import contextlib
+import logging
+import os
+import pathlib
+import signal
+import threading
 import urllib.parse
-from typing import Union, Callable, Optional
-from functools import reduce
-
-try:
-    from typing import Literal  # Added in Python 3.8
-except ImportError:
-    from typing_extensions import Literal  # type: ignore
-
-from functools import partial
+import warnings
+from functools import partial, reduce
+from typing import Any, Callable, Literal, Optional, Union
 
 import aiohttp
 from tqdm import tqdm as tqdm_std
 from tqdm.auto import tqdm as tqdm_auto
 
 import parfive
+
 from .config import DownloaderConfig, SessionConfig
 from .results import Results
 from .utils import (
+    ChecksumMismatch,
     FailedDownload,
     MultiPartDownloadError,
     Token,
     _QueueList,
     cancel_task,
+    check_file_hash,
     default_name,
     get_filepath,
     get_ftp_size,
     get_http_size,
     remove_file,
     run_task_in_thread,
+    session_head_or_get,
+    validate_checksum_format,
 )
 
 try:
@@ -93,15 +91,13 @@ class Downloader:
         self.tqdm = tqdm_auto
         if self.config.notebook is not None:
             if self.config.notebook is True:
-                from tqdm.notebook import tqdm as tqdm_notebook
+                from tqdm.notebook import tqdm as tqdm_notebook  # noqa: PLC0415
 
                 self.tqdm = tqdm_notebook
             elif self.config.notebook is False:
                 self.tqdm = tqdm_std
             else:
-                raise ValueError(
-                    "The notebook keyword argument should be one of None, True or False."
-                )
+                raise ValueError("The notebook keyword argument should be one of None, True or False.")
 
         self._configure_logging()
 
@@ -152,11 +148,10 @@ class Downloader:
     def enqueue_file(
         self,
         url: str,
-        path: Optional[Union[str, os.PathLike]] = None,
-        filename: Optional[
-            Union[str, Callable[[str, Optional[aiohttp.ClientResponse]], os.PathLike]]
-        ] = None,
-        overwrite: Optional[Union[bool, Literal["unique"]]] = None,
+        path: Union[str, os.PathLike, None] = None,
+        filename: Union[str, Callable[[str, Optional[aiohttp.ClientResponse]], os.PathLike], None] = None,
+        overwrite: Union[bool, Literal["unique"], None] = None,
+        checksum: Union[str, bool, None] = None,
         **kwargs,
     ):
         """
@@ -176,21 +171,49 @@ class Downloader:
             response will be ``None``.) If `None` the HTTP headers will be read
             for the filename, or the last segment of the URL will be used.
         overwrite
-            Determine how to handle downloading if a file already exists with the
-            same name. If `False` the file download will be skipped and the path
-            returned to the existing file, if `True` the file will be downloaded
-            and the existing file will be overwritten, if `'unique'` the filename
-            will be modified to be unique. If `None` the value set when
-            constructing the `~parfive.Downloader` object will be used.
+            Determine how to handle downloading if a file already
+            exists with the same name. If `False` the file download
+            will be skipped and the path returned to the existing file
+            (if any checksum also matches, see below), if `True` the
+            file will be downloaded and the existing file will be
+            overwritten, if `'unique'` the filename will be modified
+            to be unique. If `None` the value set when constructing
+            the `~parfive.Downloader` object will be used.
+        checksum
+            Provide a checksum, or request one from the server, to
+            compare to any existing file and verify the downloaded
+            file.
+            This option can either be a string or `True`, if `True`
+            the checksum provided by the server (if any) will be used,
+            and if it is a string the it should be of the form
+            ``<algorithm>=<checksum>``.
+            Valid algorithms are any provided by `hashlib` but the
+            HTTP spec allows ``sha-512``, ``sha-256``, ``md5`` or
+            ``sha``.
+            If ``True`` then where provided by the server (in the
+            ``Repr-Digest`` or ``Content-Digest`` headers) the
+            checksum will be validated against the checksum returned
+            by the server.
+            When ``overwrite=False`` and ``checksum=`` is set the
+            checksum will also be used to determine if any local file
+            should be overwritten.
+            If the checksum is `True` then the server provided
+            checksum will be used to compare to the local file and
+            download skipped if it matches.
+            If the checksum is explicitly passed, the download will
+            fail early if the server provides a checksum for the file
+            which doesn't match the one provided by the user.
         kwargs : `dict`
             Extra keyword arguments are passed to `aiohttp.ClientSession.request`
             or `aioftp.Client.context` depending on the protocol.
         """
         overwrite = overwrite or self.config.overwrite
+        if isinstance(checksum, str):
+            validate_checksum_format(checksum)
 
         if path is None and filename is None:
             raise ValueError("Either path or filename must be specified.")
-        elif path is None:
+        if path is None:
             path = "./"
 
         path = pathlib.Path(path)
@@ -208,7 +231,12 @@ class Downloader:
 
         if scheme in ("http", "https"):
             get_file = partial(
-                self._get_http, url=url, filepath_partial=filepath, overwrite=overwrite, **kwargs
+                self._get_http,
+                url=url,
+                filepath_partial=filepath,
+                overwrite=overwrite,
+                checksum=checksum,
+                **kwargs,
             )
             self.http_queue.append(get_file)
         elif scheme == "ftp":
@@ -289,9 +317,7 @@ class Downloader:
                     task.cancel()
                 dl_results = await asyncio.gather(*tasks, return_exceptions=True)
 
-            finally:
-                results_obj = self._format_results(dl_results, main_pb)
-                return results_obj
+            return self._format_results(dl_results, main_pb)
 
     def _format_results(self, retvals, main_pb):
         # Squash all nested lists into a single flat list
@@ -313,13 +339,12 @@ class Downloader:
         for res in retvals:
             if isinstance(res, FailedDownload):
                 results.add_error(res.filepath_partial, res.url, res.exception)
-                parfive.log.info(
-                    "%s failed to download with exception\n" "%s", res.url, res.exception
-                )
+                parfive.log.info("%s failed to download with exception\n%s", res.url, res.exception)
             elif isinstance(res, Exception):
                 raise res
             else:
-                results.append(res)
+                requested_url, filepath = res
+                results.append(path=filepath, url=requested_url)
 
         return results
 
@@ -368,8 +393,10 @@ class Downloader:
 
         new_res = self.download()
 
+        # Append paths and urls overwrite errors
         results += new_res
-        results._errors = new_res._errors
+        results._urls += new_res.urls
+        results._errors = new_res.errors
 
         return results
 
@@ -407,8 +434,7 @@ class Downloader:
         """
         if self.config.progress:
             return self.tqdm(total=total, unit="file", desc="Files Downloaded", position=0)
-        else:
-            return contextlib.contextmanager(lambda: iter([None]))()
+        return contextlib.contextmanager(lambda: iter([None]))()
 
     async def _run_http_download(self, main_pb):
         async with self.config.aiohttp_client_session() as session:
@@ -421,7 +447,7 @@ class Downloader:
 
             try:
                 # Wait for all the coroutines to finish
-                done, _ = await asyncio.wait(futures)
+                _done, _ = await asyncio.wait(futures)
             except asyncio.CancelledError:
                 for task in futures:
                     task.cancel()
@@ -437,7 +463,7 @@ class Downloader:
 
         try:
             # Wait for all the coroutines to finish
-            done, _ = await asyncio.wait(futures)
+            _done, _ = await asyncio.wait(futures)
         except asyncio.CancelledError:
             for task in futures:
                 task.cancel()
@@ -473,45 +499,56 @@ class Downloader:
 
     async def _get_http(
         self,
-        session,
+        session: aiohttp.ClientSession,
         *,
-        url,
-        filepath_partial,
-        chunksize=None,
-        file_pb=None,
-        token,
-        overwrite,
-        max_splits=None,
-        **kwargs,
-    ):
+        url: str,
+        filepath_partial: Callable,
+        chunksize: Union[int, None] = None,
+        file_pb: Union[tqdm_std, bool, None] = None,
+        token: Token,
+        overwrite: Union[bool, Literal["unique"]],
+        max_splits: Union[int, None] = None,
+        checksum: Union[str, bool, None] = None,
+        **kwargs: dict[str, Any],
+    ) -> tuple[str, str]:
         """
         Read the file from the given url into the filename given by ``filepath_partial``.
 
         Parameters
         ----------
-        session : `aiohttp.ClientSession`
+        session
             The `aiohttp.ClientSession` to use to retrieve the files.
-        url : `str`
+        url
             The url to retrieve.
-        filepath_partial : `callable`
+        filepath_partial
             A function to call which returns the filepath to save the url to.
             Takes two arguments ``resp, url``.
-        chunksize : `int`
+        chunksize
             The number of bytes to read into the file at a time.
-        file_pb : `tqdm.tqdm` or `False`
+        file_pb
             Should progress bars be displayed for each file downloaded.
-        token : `parfive.downloader.Token`
+        token
             A token for this download slot.
-        overwrite : `bool`
+        overwrite
             Overwrite the file if it already exists.
-        max_splits: `int`, optional
+        max_splits
             Number of maximum concurrent connections per file.
-        kwargs : `dict`
+        checksum
+            If given the downloaded file will be verified against the
+            given checksum.  The format of the checksum string should
+            be ``<algorithm>=<checksum>``, valid algorithms are
+            ``sha-512``, ``sha-256``, ``md5`` or ``sha``.  If ``True``
+            then where provided by the server (in the ``Repr-Digest``
+            or ``Content-Digest`` headers) the checksum will be
+            validated against the checksum returned by the server.
+        kwargs
             Extra keyword arguments are passed to `aiohttp.ClientSession.get`.
 
         Returns
         -------
-        `str`
+        url
+            The URL downloaded
+        path
             The name of the file saved.
         """
         if chunksize is None:
@@ -521,7 +558,7 @@ class Downloader:
 
         # Define filepath and writer here as we use them in the except block
         filepath = writer = None
-        tasks = []
+        tasks: list[asyncio.Task] = []
         try:
             scheme = urllib.parse.urlparse(url).scheme
             if scheme == "http":
@@ -529,7 +566,30 @@ class Downloader:
             elif scheme == "https":
                 kwargs["proxy"] = self.config.https_proxy
 
-            async with session.get(url, timeout=self.config.timeouts, **kwargs) as resp:
+            request_headers = {}
+            if checksum:
+                checksum_priority = {
+                    "sha-512": 8,
+                    "sha-256": 7,
+                    "sha": 3,
+                    "md5": 2,
+                }
+                # If the user has specified a checksum format as the
+                # server for that one with highest priority.
+                if isinstance(checksum, str):
+                    user_alg = checksum.split("=")[0]
+                    checksum_priority[user_alg] = 10
+
+                # As we are doing a HEAD / GET request for the whole
+                # file, the Repr-Digest and Content-Digest headers
+                # should be the same, so use both.
+                request_headers["Want-Repr-Digest"] = request_headers["Want-Content-Digest"] = ", ".join(
+                    [f"{k}={v}" for k, v in checksum_priority.items()]
+                )
+
+            async with session_head_or_get(
+                session, url, headers=request_headers, timeout=self.config.timeouts, **kwargs
+            ) as resp:
                 parfive.log.debug(
                     "%s request made to %s with headers=%s",
                     resp.request_info.method,
@@ -544,85 +604,136 @@ class Downloader:
                 )
                 if resp.status < 200 or resp.status >= 300:
                     raise FailedDownload(filepath_partial, url, resp)
-                else:
-                    filepath, skip = get_filepath(filepath_partial(resp, url), overwrite)
-                    if skip:
+                filepath, use_local_file_if_possible = get_filepath(filepath_partial(resp, url), overwrite)
+                # Get the expected checksum from the headers
+                header_checksum: Union[str, None] = resp.headers.get(
+                    "Repr-Digest", resp.headers.get("Content-Digest", None)
+                )
+                if checksum is True:
+                    checksum = header_checksum
+                    if checksum is None:
+                        parfive.log.info(
+                            "Expected server to provide checksum for url '%s' but none returned.", url
+                        )
+                if use_local_file_if_possible:
+                    if isinstance(checksum, str):
+                        with filepath.open(mode="rb") as fobj:
+                            checksum_matches = check_file_hash(fobj, checksum, accept_invalid_checksum=True)
+                        if checksum_matches:
+                            parfive.log.debug(
+                                "File %s already exists, checksum matches and overwrite is False; skipping download.",
+                                filepath,
+                            )
+                            return url, str(filepath)
+                    else:
                         parfive.log.debug(
                             "File %s already exists and overwrite is False; skipping download.",
                             filepath,
                         )
-                        return str(filepath)
-                    if callable(file_pb):
-                        file_pb = file_pb(
-                            position=token.n,
-                            unit="B",
-                            unit_scale=True,
-                            desc=filepath.name,
-                            leave=False,
-                            total=get_http_size(resp),
-                        )
-                    else:
-                        file_pb = None
+                        return url, str(filepath)
 
-                    # This queue will contain the downloaded chunks and their offsets
-                    # as tuples: (offset, chunk)
-                    downloaded_chunk_queue = asyncio.Queue()
-
-                    writer = asyncio.create_task(
-                        self._write_worker(downloaded_chunk_queue, file_pb, filepath)
-                    )
-
-                    if (
-                        not self.config.env.disable_range
-                        and max_splits
-                        and resp.headers.get("Accept-Ranges", None) == "bytes"
-                        and "Content-length" in resp.headers
-                    ):
-                        content_length = int(resp.headers["Content-length"])
-                        split_length = max(1, content_length // max_splits)
-                        ranges = [
-                            [start, start + split_length]
-                            for start in range(0, content_length, split_length)
-                        ]
-                        # let the last part download everything
-                        ranges[-1][1] = ""
-                        for _range in ranges:
-                            tasks.append(
-                                asyncio.create_task(
-                                    self._http_download_worker(
-                                        session,
-                                        url,
-                                        chunksize,
-                                        _range,
-                                        downloaded_chunk_queue,
-                                        **kwargs,
-                                    )
-                                )
-                            )
-                    else:
-                        tasks.append(
-                            asyncio.create_task(
-                                self._http_download_worker(
-                                    session,
+                if isinstance(checksum, str) and header_checksum is not None:
+                    try:
+                        header_alg, header_chk = validate_checksum_format(header_checksum)
+                        user_alg, user_chk = validate_checksum_format(checksum)
+                        if header_alg == user_alg:
+                            if header_chk != user_chk:
+                                raise FailedDownload(
+                                    filepath,
                                     url,
-                                    chunksize,
-                                    None,
-                                    downloaded_chunk_queue,
-                                    **kwargs,
+                                    ChecksumMismatch(
+                                        "Server provided checksum and user provided checksum do not match, download skipped"
+                                    ),
                                 )
+                        else:
+                            parfive.log.info(
+                                "Not comparing user provided checksum to server provided checksum"
+                                " as algorithms do not match (got %s from the server).",
+                                header_alg,
+                            )
+                    except ValueError as e:
+                        parfive.log.info(
+                            "Failed to compare user checksum to server checksum due to error: %s", e
+                        )
+
+                if callable(file_pb):
+                    file_pb = file_pb(
+                        position=token.n,
+                        unit="B",
+                        unit_scale=True,
+                        desc=filepath.name,
+                        leave=False,
+                        total=get_http_size(resp),
+                    )
+                else:
+                    file_pb = None
+
+                # This queue will contain the downloaded chunks and their offsets
+                # as tuples: (offset, chunk)
+                downloaded_chunk_queue: asyncio.Queue = asyncio.Queue()
+
+                writer = asyncio.create_task(self._write_worker(downloaded_chunk_queue, file_pb, filepath))
+
+                if (
+                    not self.config.env.disable_range
+                    and max_splits
+                    and resp.headers.get("Accept-Ranges", None) == "bytes"
+                    and "Content-length" in resp.headers
+                ):
+                    content_length = int(resp.headers["Content-length"])
+                    split_length = max(1, content_length // max_splits)
+                    ranges: list[list[Union[int, str]]] = [
+                        [start, start + split_length] for start in range(0, content_length, split_length)
+                    ]
+                    # let the last part download everything
+                    ranges[-1][1] = ""
+                    tasks.extend(
+                        asyncio.create_task(
+                            self._http_download_worker(
+                                session,
+                                url,
+                                chunksize,
+                                _range,
+                                downloaded_chunk_queue,
+                                **kwargs,
                             )
                         )
+                        for _range in ranges
+                    )
+                else:
+                    tasks.append(
+                        asyncio.create_task(
+                            self._http_download_worker(
+                                session,
+                                url,
+                                chunksize,
+                                None,
+                                downloaded_chunk_queue,
+                                **kwargs,
+                            )
+                        )
+                    )
             # Close the initial request here before we start transferring data.
 
             # run all the download workers
             await asyncio.gather(*tasks)
             # join() waits till all the items in the queue have been processed
             await downloaded_chunk_queue.join()
+
+            with filepath.open(mode="rb") as fobj:
+                if isinstance(checksum, str) and not check_file_hash(
+                    fobj, checksum, accept_invalid_checksum=True
+                ):
+                    raise FailedDownload(
+                        filepath, url, ChecksumMismatch("Downloaded checksum doesn't match.")
+                    )
+
             for callback in self.config.done_callbacks:
                 callback(filepath, url, None)
-            return str(filepath)
 
-        except (Exception, asyncio.CancelledError) as e:
+            return url, str(filepath)
+
+        except (Exception, asyncio.CancelledError) as e:  # noqa: BLE001
             for task in tasks:
                 task.cancel()
             # We have to cancel the writer here before we try and remove the
@@ -668,7 +779,7 @@ class Downloader:
             await self._blocking_write_worker(queue, file_pb, filepath)
 
     async def _async_write_worker(self, queue, file_pb, filepath):
-        import aiofiles
+        import aiofiles  # noqa: PLC0415
 
         async with aiofiles.open(filepath, mode="wb") as f:
             while True:
@@ -798,9 +909,7 @@ class Downloader:
             async with aioftp.Client.context(parse.hostname, **kwargs) as client:
                 parfive.log.debug("Connected to ftp server %s", parse.hostname)
                 if parse.username and parse.password:
-                    parfive.log.debug(
-                        "Explicitly Logging in with %s:%s", parse.username, parse.password
-                    )
+                    parfive.log.debug("Explicitly Logging in with %s:%s", parse.username, parse.password)
                     await client.login(parse.username, parse.password)
 
                 # This has to be done before we start streaming the file:
@@ -810,7 +919,7 @@ class Downloader:
                         "File %s already exists and overwrite is False; skipping download.",
                         filepath,
                     )
-                    return str(filepath)
+                    return url, str(filepath)
 
                 if callable(file_pb):
                     total_size = await get_ftp_size(client, parse.path)
@@ -834,9 +943,7 @@ class Downloader:
                     )
 
                     download_workers.append(
-                        asyncio.create_task(
-                            self._ftp_download_worker(stream, downloaded_chunks_queue)
-                        )
+                        asyncio.create_task(self._ftp_download_worker(stream, downloaded_chunks_queue))
                     )
 
                     await asyncio.gather(*download_workers)
@@ -845,9 +952,9 @@ class Downloader:
                     for callback in self.config.done_callbacks:
                         callback(filepath, url, None)
 
-                    return str(filepath)
+                    return url, str(filepath)
 
-        except (Exception, asyncio.CancelledError) as e:
+        except (Exception, asyncio.CancelledError) as e:  # noqa: BLE001
             if writer is not None:
                 await cancel_task(writer)
                 writer = None
